/* sim_console.c: simulator cmd I/O library

   Copyright (c) 1993-2006, Robert M Supnik

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   ROBERT M SUPNIK BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of Robert M Supnik shall not be
   used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from Robert M Supnik.

   Added by Joerg Hoppe, Goettingen, Germany. j_hoppe@t-online.de

   30-Jun-06    JH      cloned from sim_console.c

   This module implements the following routines to support stdioio over telnet:

   simf_fpriiintf       central stream output, routes any output to
                        stream, stderr, stdio or telent
   sim_fpuuuts          dto. for "fputs" output style

//JH   sim_poll_kbd -       poll for keyboard input
//JH  sim_putchar  -       output character to console
//JH  sim_putchar_s -      output character to console, stall if congested
//JH  sim_set_console -    set console parameters
//JH  sim_show_console -   show console parameters
//JH  sim_tt_inpcvt -      convert input character per mode
//JH  sim_tt_outcvt -      convert output character per mode
//JH
//JH  sim_ttinit   -       called once to get initial terminal state
//JH  sim_ttrun    -       called to put terminal into run state
//JH  sim_ttcmd    -       called to return terminal to command state
//JH  sim_ttclose  -       called once before the simulator exits
//JH  sim_os_poll_kbd -    poll for keyboard input
//JH  sim_os_putchar -     output character to console

   The first group is OS-independent; the second group is OS-dependent.

   The following routines are exposed but deprecated:

   sim_set_telnet -     set console to Telnet port
   sim_set_notelnet -   close console Telnet port
   sim_show_telnet -    show console status
*/

#include "sim_defs.h"
#include <stdio.h>
#include <assert.h>
//#include <varargs.h>
#include <stdarg.h>                    // var arg

#include "sim_sock.h"
#include "sim_tmxr.h"


// aus sim_console.c
extern int sim_del_char ;
extern int sim_int_char ;
extern int sim_brk_char ;


//JH #include <ctype.h>
//JH
//JH #define KMAP_WRU        0
//JH #define KMAP_BRK        1
//JH #define KMAP_DEL        2
//JH #define KMAP_MASK       0377
//JH #define KMAP_NZ         0400
//JH
//JH int32 sim_int_char = 005;                               /* interrupt character */
//JH int32 sim_brk_char = 000;                               /* break character */
//JH int32 sim_tt_pchar = 0x00002780;
//JH #if defined (_WIN32) || defined (__OS2__) || (defined (__MWERKS__) && defined (macintosh))
//JH int32 sim_del_char = '\b';                              /* delete character */
//JH #else
//JH int32 sim_del_char = 0177;
//JH #endif
// telnet connection is copied form temrinal multiplexer logic
// only one line is used: master
TMLN sim_stdio_ldsc = { 0 };                              /* console line descr */
TMXR sim_stdio_tmxr = { 1, 0, 0, &sim_stdio_ldsc };         /* console line mux */
//JH
//JH extern volatile int32 stop_cpu;
//JH extern int32 sim_quiet, sim_deb_close;
//JH extern FILE *sim_log, *sim_deb;
//JH extern DEVICE *sim_devices[];
//JH
//JH /* Set/show data structures */
//JH
static CTAB set_stdio_tab[] = {
//JH     { "WRU", &sim_set_kmap, KMAP_WRU | KMAP_NZ },
//JH     { "BRK", &sim_set_kmap, KMAP_BRK },
//JH     { "DEL", &sim_set_kmap, KMAP_DEL |KMAP_NZ },
//JH     { "PCHAR", &sim_set_pchar, 0 },
    { "TELNET", &sim_stdio_set_telnet, 0 },
    { "NOTELNET", &sim_stdio_set_notelnet, 0 },
//JH     { "LOG", &sim_set_logon, 0 },
//JH     { "NOLOG", &sim_set_logoff, 0 },
//JH     { "DEBUG", &sim_set_debon, 0 },
//JH     { "NODEBUG", &sim_set_deboff, 0 },
     { NULL, NULL, 0 }
     };
//JH
 static SHTAB show_stdio_tab[] = {
//JH     { "WRU", &sim_show_kmap, KMAP_WRU },
//JH     { "BRK", &sim_show_kmap, KMAP_BRK },
//JH     { "DEL", &sim_show_kmap, KMAP_DEL },
//JH     { "PCHAR", &sim_show_pchar, 0 },
//JH     { "LOG", &sim_show_log, 0 },
     { "TELNET", &sim_stdio_show_telnet, 0 },
//JH     { "DEBUG", &sim_show_debug, 0 },
     { NULL, NULL, 0 }
     };
//JH
//JH static int32 *cons_kmap[] = {
//JH     &sim_int_char,
//JH     &sim_brk_char,
//JH     &sim_del_char
//JH     };
//JH

 // input buffer
#define SIM_STDIO_TELNET_INPUTLINE_SIZE 256
static char sim_stdio_telnet_input_buff[SIM_STDIO_TELNET_INPUTLINE_SIZE+1] ;
//static char sim_stdio_telnet_last_inputchar = 0 ;
static char *sim_stdio_telnet_input_writeptr ;
//static int sim_stdio_telnet_inputline_ready = 0 ;

// all output to former "printf()" and sim_fpriiintf(stderr, ) is routed through this!
// routes user-visible output over cmd_telnet
// also no "puts()" or "sim_fpuuuts()" is allowed any more.
// JH

 // must be called periodically.
 // performs background char receiving for sim_fgeets()
 // Buffering is only necessary to make
 // a char preview with sim_stdio_input_getc(1)
 // Telnet is only polled if buffer is (almost) empty
 t_stat sim_stdio_input_poll(void)
 {
	 if (sim_stdio_tmxr.master != 0)  {
		 // only functional, if stdin over telnet
		 int ready = 0 ; // 1, if telnet inut  buffer empty
    int32 c;
//{ char buff[100] ; sprintf(buff, "sim_stdio_poll_input(): ENTER\n") ; OutputDebugStringA(buff) ; }

	if (sim_stdio_ldsc.conn == 0)
		  return SCPE_LOST;           /* no Telnet conn? */
	if (!sim_stdio_input_fill()) {
//	while(!ready) {
      tmxr_poll_rx (&sim_stdio_tmxr);                           /* poll for input */
	  ready = 1 ;
	  if (c = tmxr_getc_ln (&sim_stdio_ldsc)) {                   /* any char? */
//{ char buff[100] ; sprintf(buff, "sim_stdio_poll_input(): rcvd '%c'\n", c) ; OutputDebugStringA(buff) ;}
		  ready = 0 ; // loop again
		  c &= 0377 ;

			// first, assert valid writeptr
			if (!sim_stdio_telnet_input_writeptr)
				sim_stdio_telnet_input_writeptr = sim_stdio_telnet_input_buff ;
		// Zeichen in Puffer
		  		if (sim_stdio_telnet_input_writeptr - sim_stdio_telnet_input_buff < SIM_STDIO_TELNET_INPUTLINE_SIZE)
					*sim_stdio_telnet_input_writeptr++ = c ;
				else return SCPE_TTIERR ; // buffer voll, char lost: right error code?
				*sim_stdio_telnet_input_writeptr = '\0' ;
		}
/***
		  sim_stdio_telnet_last_inputchar = c ; // letztes zeichen merken
//NEIN!!! in fgeets visible!!! Denn poll_kbd echot nicht!
		  //echo c, even backspace and \n
{ char buff[100] ; sprintf(buff, "sim_stdio_poll_input(): rcvd '%c'\n", c) ; OutputDebugStringA(buff) ;}
			sim_fpriiintf(stdout, "%c", c) ;
			// first, assert valid writeptr
			if (!sim_stdio_telnet_inputline_writeptr)
				sim_stdio_telnet_inputline_writeptr = sim_stdio_telnet_inputline_buff ;
			switch(c) {
			case 0x0d: // CR
  			// line ready
				sim_stdio_telnet_inputline_ready = 1 ; // signal to
				// next char begins new input line, even if line not read by 'sim_fgeeets()'
				sim_stdio_telnet_inputline_writeptr = NULL ;
				return SCPE_OK ;
				break ;
			case 0x08: // backspace: erase last char from buffer
				if (sim_stdio_telnet_inputline_writeptr > sim_stdio_telnet_inputline_buff)
					sim_stdio_telnet_inputline_writeptr-- ;
				*sim_stdio_telnet_inputline_writeptr = '\0' ;
				break ;
			default:
				// add char to buffer
				if (sim_stdio_telnet_inputline_writeptr - sim_stdio_telnet_inputline_buff < SIM_STDIO_TELNET_INPUTLINE_SIZE)
					*sim_stdio_telnet_inputline_writeptr++ = c ;
				else return SCPE_TTIERR ; // buffer voll, char lost: right error code?
				*sim_stdio_telnet_inputline_writeptr = '\0' ;
			}
***/
	  }
	 }
	 return SCPE_OK ;
 }

 int sim_stdio_input_fill(void) {
			if (!sim_stdio_telnet_input_writeptr)
				sim_stdio_telnet_input_writeptr = sim_stdio_telnet_input_buff ;

			return sim_stdio_telnet_input_writeptr - sim_stdio_telnet_input_buff ;
 }

 // ltestes Zeichen aus Buffer holen
 // 'inspect_only' = 1: nicht aus Buffer entfernen
int sim_stdio_input_getc(int inspect_only) {


	int	c ;
	int	n ;
	t_stat	r ;
	if (!sim_stdio_telnet_input_writeptr)
			sim_stdio_telnet_input_writeptr = sim_stdio_telnet_input_buff ;

	if ((r= sim_stdio_input_poll()) != SCPE_OK)
		return r ; // LOST?
	n = sim_stdio_input_fill() ;
	if (!n)
		return EOF ;
	else {
		c = *sim_stdio_telnet_input_buff ; // ltestes Zeichen zurckgeben
		if (! inspect_only) {
			// Zeichen aus Buffer entfernen
			memmove(sim_stdio_telnet_input_buff, sim_stdio_telnet_input_buff+1, n-1) ;
			sim_stdio_telnet_input_writeptr-- ;
			assert(sim_stdio_telnet_input_writeptr >= sim_stdio_telnet_input_buff) ;
		}
		return c ;
	}
}



 t_stat sim_os_polllll_kbd (void)
 {
		int32 c	 ;
		t_stat	r ;
	 if (sim_stdio_tmxr.master != 0) {
	if ((r= sim_stdio_input_poll()) != SCPE_OK)
		return r ; // LOST?

		 if (!sim_stdio_input_fill())
			 return SCPE_OK ; // no char
		 else {
			 c = sim_stdio_input_getc(0) ;
		 if ((c & 0177) == sim_del_char) c = 0177;
		 if ((c & 0177) == sim_int_char) return SCPE_STOP;
		 if (sim_brk_char && ((c & 0177) == sim_brk_char)) return SCPE_BREAK;
//{ char buff[100] ; sprintf(buff, "sim_os_polllll_kbd(): '%c'\n", c) ; OutputDebugStringA(buff) ;}
		 return c | SCPE_KFLAG;
		 }
/***
    	if (sim_stdio_ldsc.conn == 0)
	 	  return SCPE_LOST;           // no Telnet conn?
		  tmxr_poll_rx (&sim_stdio_tmxr);                           // poll for input
	  if (c = tmxr_getc_ln (&sim_stdio_ldsc)) {                   // any char?
		 if ((c & 0177) == sim_del_char) c = 0177;
		 if ((c & 0177) == sim_int_char) return SCPE_STOP;
		 if (sim_brk_char && ((c & 0177) == sim_brk_char)) return SCPE_BREAK;
{ char buff[100] ; sprintf(buff, "sim_os_polllll_kbd(): '%c'\n", c) ; OutputDebugStringA(buff) ;}
		 return c | SCPE_KFLAG;
	  }
		 else return SCPE_OK ;
***

		 int c;
		 sim_stdio_poll_input() ; // receive all pending telnet input
		 // wenn was im input buffer steht: es gilt als tastendruck

		 if (!sim_stdio_telnet_last_inputchar)
			return SCPE_OK;
		 // c = _getch() ;
		 c = sim_stdio_telnet_last_inputchar ;
		 sim_stdio_telnet_last_inputchar = 0 ;
		 // die ganze eingabe lschen
		 *sim_stdio_telnet_inputline_buff = '\0' ;

		 if ((c & 0177) == sim_del_char) c = 0177;
		 if ((c & 0177) == sim_int_char) return SCPE_STOP;
		 if (sim_brk_char && ((c & 0177) == sim_brk_char)) return SCPE_BREAK;
		 return c | SCPE_KFLAG;
		 ***/
 } else
 	return sim_os_poll_kbd() ; // keine Umleitung nach telnet
 }

 // fgets() auf Telnet buffer
char * sim_fgeeets (char *cptr, int size, FILE *stream)
{
 if (sim_stdio_tmxr.master != 0 && stream == stdin) {
		int c ;
		char	*wp = cptr ;
		t_stat	r ;

		// Schleife, bis CR empfangen wurde
		while(1) {
	  	 if (sim_check_stdio(10) != SCPE_OK)  // check and reconnect
				return NULL ; // telnet lost!

		 c = sim_stdio_input_getc(0) ;
		 if (c == EOF) {
			 // no new char: wait 1ms and retry
			sim_os_ms_sleep(1) ;
		 } else {
			sim_fpriiintf(stdout, "%c", c) ; // ECHO

			switch(c) {
			case 0x0d: // CR
  			// line ready
				*wp = NULL ;
				return cptr ;
			case 0x08: // backspace: erase last char from buffer
				if (wp > cptr)
					wp-- ;
				break ;
			default:
				// add char to buffer
				*wp++ = c ;
				if ((wp - cptr ) >= SIM_STDIO_TELNET_INPUTLINE_SIZE) {
					*wp = NULL ;
					return cptr ; // Emergency exit
				}
			}
			}
		}
/***


	 // receive string over telnet:
	 while (!sim_stdio_telnet_inputline_ready) {
    	 if (sim_check_stdio(10) != SCPE_OK)  // check and reconnect
				return NULL ; // telnet lost!
		 sim_stdio_poll_input() ; // receive all pending telnet input
	     sim_os_ms_sleep(10) ; // don't heat cpu while waiting
    }
	// input line received in background?
	 if (sim_stdio_telnet_inputline_ready) {
		// copy string to cptr
		 int n = strlen(sim_stdio_telnet_inputline_buff) ;
		 if (n > size-1)  // cptr too small, truncate input
		 	 n = size-1 ;
		   strncpy(cptr, sim_stdio_telnet_inputline_buff, n) ;
		   cptr[n] = '\0' ;
		 sim_stdio_telnet_inputline_ready = 0 ;

			 // hier erst das echo machen!!!
		// sim_fpriiintf(stdout, "%s", cptr) ;
		 OutputDebugStringA("sim_fgeeets(): ") ;OutputDebugStringA(cptr) ;OutputDebugStringA("\n") ;
		 return cptr ;
	 } else
		 return NULL ;
		 ***/
 } else
		return fgets(cptr, size, stream) ;
}


int sim_fpriiintf(FILE *stream, char *fmt, ...)
{
		char	buff[2048] ; // fixed size: very bad idea .....
		int i ;
	    va_list args;
		va_start(args, fmt);

		vsnprintf(buff, sizeof(buff), fmt, args) ;
	 if (sim_stdio_tmxr.master != 0 && (stream == stdout || stream == stderr)) {
		 for (i=0 ; buff[i] ; i++)
			sim_stdio_putchar(buff[i]) ;
// 	 return sim_write_sock (sim_stdio_tmxr.master, buff, strlen(buff)) ;
	 	// -1 = failure
		 return i ; // number of bytes
     } else
		return fprintf(stream, buff) ;
} ;

int sim_fpuuuts(char *str, FILE *stream) {
//int32 sim_write_sock (SOCKET sock, char *msg, int32 nbytes)

	return sim_fpriiintf(stream, str) ;
}

int sim_fpuuutc(int c, FILE *stream)
{
if (sim_stdio_tmxr.master != 0 && (stream == stdout || stream == stderr)) {
 		sim_stdio_putchar(c) ;
		 return c ; // return char
     } else
		return fputc(c, stream) ;
}




/* stdio package.

	The stdio input/output for the controlling window can be attached
	to a Telnet connection.  If attached to a Telnet connection,
 */
//JH    the console is described by internal terminal multiplexor
//JH    sim_con_tmxr and internal terminal line description sim_con_ldsc.
//JH */
//JH
//JH /* SET STDIO command */
//JH
t_stat sim_set_stdio (int32 flag, char *cptr)
{
 char *cvptr, gbuf[CBUFSIZE];
 CTAB *ctptr;
 t_stat r;

 if ((cptr == NULL) || (*cptr == 0)) return SCPE_2FARG;
 while (*cptr != 0) {                                    /* do all mods */
     cptr = get_glyph_nc (cptr, gbuf, ',');              /* get modifier */
     if (cvptr = strchr (gbuf, '=')) *cvptr++ = 0;       /* = value? */
     get_glyph (gbuf, gbuf, 0);                          /* modifier to UC */
     if (ctptr = find_ctab (set_stdio_tab, gbuf)) {        /* match? */
         r = ctptr->action (ctptr->arg, cvptr);          /* do the rest */
         if (r != SCPE_OK) return r;
         }
     else return SCPE_NOPARAM;
     }
  return SCPE_OK;
}

 /* SHOW STDIO command */

t_stat sim_show_stdio (FILE *st, DEVICE *dptr, UNIT *uptr, int32 flag, char *cptr)
{
char gbuf[CBUFSIZE];
SHTAB *shptr;
int32 i;
if (*cptr == 0) {                                       /* show all */
    for (i = 0; show_stdio_tab[i].name; i++)
        show_stdio_tab[i].action (st, dptr, uptr, show_stdio_tab[i].arg, cptr);
    return SCPE_OK;
    }
while (*cptr != 0) {
    cptr = get_glyph (cptr, gbuf, ',');                 /* get modifier */
    if (shptr = find_shtab (show_stdio_tab, gbuf))
        shptr->action (st, dptr, uptr, shptr->arg, cptr);
    else return SCPE_NOPARAM;
    }
 return SCPE_OK;
}

//JH /* Set keyboard map */
//JH
//JH t_stat sim_set_kmap (int32 flag, char *cptr)
//JH {
//JH DEVICE *dptr = sim_devices[0];
//JH int32 val, rdx;
//JH t_stat r;
//JH
//JH if ((cptr == NULL) || (*cptr == 0)) return SCPE_2FARG;
//JH if (dptr->dradix == 16) rdx = 16;
//JH else rdx = 8;
//JH val = (int32) get_uint (cptr, rdx, 0177, &r);
//JH if ((r != SCPE_OK) ||
//JH     ((val == 0) && (flag & KMAP_NZ))) return SCPE_ARG;
//JH *(cons_kmap[flag & KMAP_MASK]) = val;
//JH return SCPE_OK;
//JH }
//JH
//JH /* Show keyboard map */
//JH
//JH t_stat sim_show_kmap (FILE *st, DEVICE *dptr, UNIT *uptr, int32 flag, char *cptr)
//JH {
//JH if (sim_devices[0]->dradix == 16)
//JH     sim_fpriiintf (st, "%s = %X\n", show_con_tab[flag].name, *(cons_kmap[flag & KMAP_MASK]));
//JH else sim_fpriiintf (st, "%s = %o\n", show_con_tab[flag].name, *(cons_kmap[flag & KMAP_MASK]));
//JH return SCPE_OK;
//JH }
//JH
//JH /* Set printable characters */
//JH
//JH t_stat sim_set_pchar (int32 flag, char *cptr)
//JH {
//JH DEVICE *dptr = sim_devices[0];
//JH uint32 val, rdx;
//JH t_stat r;
//JH
//JH if ((cptr == NULL) || (*cptr == 0)) return SCPE_2FARG;
//JH if (dptr->dradix == 16) rdx = 16;
//JH else rdx = 8;
//JH val = (uint32) get_uint (cptr, rdx, 0xFFFFFFFF, &r);
//JH if ((r != SCPE_OK) ||
//JH     ((val & 0x00002400) == 0)) return SCPE_ARG;
//JH sim_tt_pchar = val;
//JH return SCPE_OK;
//JH }
//JH
//JH /* Show printable characters */
//JH
//JH t_stat sim_show_pchar (FILE *st, DEVICE *dptr, UNIT *uptr, int32 flag, char *cptr)
//JH {
//JH if (sim_devices[0]->dradix == 16)
//JH     sim_fpriiintf (st, "pchar mask = %X\n", sim_tt_pchar);
//JH else sim_fpriiintf (st, "pchar mask = %o\n", sim_tt_pchar);
//JH return SCPE_OK;
//JH }
//JH
//JH /* Set log routine */
//JH
//JH t_stat sim_set_logon (int32 flag, char *cptr)
//JH {
//JH char gbuf[CBUFSIZE];
//JH
//JH if ((cptr == NULL) || (*cptr == 0)) return SCPE_2FARG;  /* need arg */
//JH cptr = get_glyph_nc (cptr, gbuf, 0);                    /* get file name */
//JH if (*cptr != 0) return SCPE_2MARG;                      /* now eol? */
//JH sim_set_logoff (0, NULL);                               /* close cur log */
//JH sim_log = sim_fopen (gbuf, "a");                        /* open log */
//JH if (sim_log == NULL) return SCPE_OPENERR;               /* error? */
//JH if (!sim_quiet) sim_fpriiintf (stdout, "Logging to file \"%s\"\n", gbuf);
//JH sim_fpriiintf (sim_log, "Logging to file \"%s\"\n", gbuf);    /* start of log */
//JH return SCPE_OK;
//JH }
//JH
//JH /* Set nolog routine */
//JH
//JH t_stat sim_set_logoff (int32 flag, char *cptr)
//JH {
//JH if (cptr && (*cptr != 0)) return SCPE_2MARG;            /* now eol? */
//JH if (sim_log == NULL) return SCPE_OK;                    /* no log? */
//JH if (!sim_quiet) sim_fpriiintf (stdout, "Log file closed\n");
//JH sim_fpriiintf (sim_log, "Log file closed\n");                 /* close log */
//JH fclose (sim_log);
//JH sim_log = NULL;
//JH return SCPE_OK;
//JH }
//JH
//JH /* Show log status */
//JH
//JH t_stat sim_show_log (FILE *st, DEVICE *dptr, UNIT *uptr, int32 flag, char *cptr)
//JH {
//JH if (cptr && (*cptr != 0)) return SCPE_2MARG;
//JH if (sim_log) sim_fpuuuts ("Logging enabled\n", st);
//JH else sim_fpuuuts ("Logging disabled\n", st);
//JH return SCPE_OK;
//JH }
//JH
//JH /* Set debug routine */
//JH
//JH t_stat sim_set_debon (int32 flag, char *cptr)
//JH {
//JH char *tptr, gbuf[CBUFSIZE];
//JH
//JH if ((cptr == NULL) || (*cptr == 0)) return SCPE_2FARG;  /* too few arguments? */
//JH tptr = get_glyph (cptr, gbuf, 0);                       /* get file name */
//JH if (*tptr != 0) return SCPE_2MARG;                      /* now eol? */
//JH sim_set_deboff (0, NULL);                               /* close cur debug */
//JH if (strcmp (gbuf, "LOG") == 0) {                        /* debug to log? */
//JH     if (sim_log == NULL) return SCPE_ARG;               /* any log? */
//JH     sim_deb = sim_log;
//JH     }
//JH else if (strcmp (gbuf, "STDOUT") == 0) sim_deb = stdout; /* debug to stdout? */
//JH else if (strcmp (gbuf, "STDERR") == 0) sim_deb = stderr; /* debug to stderr? */
//JH else {
//JH     cptr = get_glyph_nc (cptr, gbuf, 0);                /* reparse */
//JH     sim_deb = sim_fopen (gbuf, "a");                    /* open debug */
//JH     if (sim_deb == NULL) return SCPE_OPENERR;           /* error? */
//JH     sim_deb_close = 1;                                  /* need close */
//JH     }
//JH if (!sim_quiet) sim_fpriiintf (stdout, "Debug output to \"%s\"\n", gbuf);
//JH if (sim_log) sim_fpriiintf (sim_log, "Debug output to \"%s\"\n", gbuf);
//JH return SCPE_OK;
//JH }
//JH
//JH /* Set nodebug routine */
//JH
//JH t_stat sim_set_deboff (int32 flag, char *cptr)
//JH {
//JH if (cptr && (*cptr != 0)) return SCPE_2MARG;            /* now eol? */
//JH if (sim_deb == NULL) return SCPE_OK;                    /* no log? */
//JH if (!sim_quiet) sim_fpriiintf (stdout, "Debug output disabled\n");
//JH if (sim_log) sim_fpriiintf (sim_log, "Debug output disabled\n");
//JH if (sim_deb_close) fclose (sim_deb);                    /* close if needed */
//JH sim_deb_close = 0;
//JH sim_deb = NULL;
//JH return SCPE_OK;
//JH }
//JH
//JH /* Show debug routine */
//JH
//JH t_stat sim_show_debug (FILE *st, DEVICE *dptr, UNIT *uptr, int32 flag, char *cptr)
//JH {
//JH if (cptr && (*cptr != 0)) return SCPE_2MARG;
//JH if (sim_deb) sim_fpuuuts ("Debug output enabled\n", st);
//JH else sim_fpuuuts ("Debug output disabled\n", st);
//JH return SCPE_OK;
//JH }
//JH
/* Set console to Telnet port */

t_stat sim_stdio_set_telnet (int32 flg, char *cptr)
{
if ((cptr == NULL) || (*cptr == 0)) return SCPE_2FARG;  /* too few arguments? */
if (sim_stdio_tmxr.master) return SCPE_ALATT;             /* already open? */
tmxr_open_master (&sim_stdio_tmxr, cptr);          /* open master socket */
return sim_check_stdio(30) ; // check and reconnect

}

/* Close console Telnet port */

t_stat sim_stdio_set_notelnet (int32 flag, char *cptr)
{
if (cptr && (*cptr != 0)) return SCPE_2MARG;            /* too many arguments? */
if (sim_stdio_tmxr.master == 0) return SCPE_OK;           /* ignore if already closed */
return tmxr_close_master (&sim_stdio_tmxr);               /* close master socket */
}

 /* Show console Telnet status */

 t_stat sim_stdio_show_telnet (FILE *st, DEVICE *dunused, UNIT *uunused, int32 flag, char *cptr)
 {
 if (cptr && (*cptr != 0)) return SCPE_2MARG;
 if (sim_stdio_tmxr.master == 0)
     sim_fpriiintf (st, "Connected to command window\n");
 else if (sim_stdio_ldsc.conn == 0)
     sim_fpriiintf (st, "Listening on port %d\n", sim_stdio_tmxr.port);
 else {
     sim_fpriiintf (st, "Listening on port %d, connected to socket %d\n",
         sim_stdio_tmxr.port, sim_stdio_ldsc.conn);
     tmxr_fconns (st, &sim_stdio_ldsc, -1);
     tmxr_fstats (st, &sim_stdio_ldsc, -1);
     }
 return SCPE_OK;
 }

/* Check connection before executing */

t_stat sim_check_stdio (int32 sec)
{
int32 c, i;

if (sim_stdio_tmxr.master == 0) return SCPE_OK;           /* not Telnet? done */
if (sim_stdio_ldsc.conn) {                                /* connected? */
    tmxr_poll_rx (&sim_stdio_tmxr);                       /* poll (check disconn) */
    if (sim_stdio_ldsc.conn) return SCPE_OK;              /* still connected? */
   }
for (i = 0; i < sec; i++) {                             /* loop */
    if (tmxr_poll_conn (&sim_stdio_tmxr) >= 0) {          /* poll connect */
        sim_stdio_ldsc.rcve = 1;                          /* rcv enabled */
        if (i) {                                        /* if delayed */

            fprintf (stdout, "Running\n");                /* print transition, always to physical stdout */
            fflush (stdout);
            }
        return SCPE_OK;                                 /* ready to proceed */
        }
    c = sim_os_poll_kbd ();                             /* check for stop char */
    if ((c == SCPE_STOP)/* || stop_cpu*/) return SCPE_STOP;
    if ((i % 10) == 0) {                                /* Status every 10 sec */
        fprintf (stdout, "Waiting for stdio telnet connection\n");
        fflush (stdout);
        }
    sim_os_sleep (1);                                   /* wait 1 second */
    }
	// close connection.
	tmxr_close_master (&sim_stdio_tmxr);
    fprintf (stdout, "Timeout: closing stdio telnet connection\n");
    fflush (stdout);
	return SCPE_TTMO;                                       /* timed out */
}



/* Output character */
t_stat sim_stdio_putchar (int32 c)
{
//if (sim_log) fputc (c, sim_log);                        /* log file? */
if (sim_stdio_tmxr.master == 0)                           /* not Telnet? */
    return sim_os_putchar (c);                          /* in-window version */
if (sim_stdio_ldsc.conn == 0) return SCPE_LOST;           /* no Telnet conn? */
tmxr_putc_ln (&sim_stdio_ldsc, c);                        /* output char */
tmxr_poll_tx (&sim_stdio_tmxr);                           /* poll xmt */
return SCPE_OK;
}


/* Stdio input service
 */

t_stat sim_stdio_svc (void)
{
 t_stat stat ;
 int32 c ;
 extern volatile int32 stop_cpu;
 extern int32 sim_int_char;

	// restore telnet connection
	stat = sim_check_stdio(10) ;
	if (stat != SCPE_OK)
       return stat ;

 // monitor telnet input for break
 if (sim_stdio_ldsc.conn == 0)
   return SCPE_LOST;		   /* no Telnet conn? */

	// nondestrcutive keybd poll
	if (c =  sim_stdio_input_getc(/* inspect_only*/1))
		   // eval special key codes
		   // code is like "sim_poll_kbd ()"
    		//	 if ((c & 0177) == sim_del_char) c = 0177;
		   if ((c & 0177) == sim_int_char) {	/* ^E ? */
			   stop_cpu = 1;
	}

 /*** tmxr_poll_rx (&sim_stdio_tmxr);							 // poll for input
if (c = tmxr_getc_ln (&sim_stdio_ldsc)) {				   // any char?
   // eval special key codes
   // code is like "sim_poll_kbd ()"
//   if ((c & 0177) == sim_del_char) c = 0177;
   if ((c & 0177) == sim_int_char) { 	// ^E ?
	   stop_cpu = 1;
   	}
//   if (sim_brk_char && ((c & 0177) == sim_brk_char)) return SCPE_BREAK;


  }
   ***/
return SCPE_OK;
}



//JH t_stat sim_putchar_s (int32 c)
//JH {
//JH t_stat r;
//JH
//JH if (sim_log) sim_fpuuutc (c, sim_log);                        /* log file? */
//JH if (sim_con_tmxr.master == 0)                           /* not Telnet? */
//JH     return sim_os_putchar (c);                          /* in-window version */
//JH if (sim_con_ldsc.conn == 0) return SCPE_LOST;           /* no Telnet conn? */
//JH if (sim_con_ldsc.xmte == 0) r = SCPE_STALL;             /* xmt disabled? */
//JH else r = tmxr_putc_ln (&sim_con_ldsc, c);               /* no, Telnet output */
//JH tmxr_poll_tx (&sim_con_tmxr);                           /* poll xmt */
//JH return r;                                               /* return status */
//JH }
//JH
//JH /* Input character processing */
//JH
//JH int32 sim_tt_inpcvt (int32 c, uint32 mode)
//JH {
//JH uint32 md = mode & TTUF_M_MODE;
//JH
//JH if (md != TTUF_MODE_8B) {
//JH     c = c & 0177;
//JH     if (md == TTUF_MODE_UC) {
//JH         if (islower (c)) c = toupper (c);
//JH         if (mode & TTUF_KSR) c = c | 0200;
//JH         }
//JH     }
//JH else c = c & 0377;
//JH return c;
//JH }
//JH
//JH /* Output character processing */
//JH
//JH int32 sim_tt_outcvt (int32 c, uint32 mode)
//JH {
//JH uint32 md = mode & TTUF_M_MODE;
//JH
//JH if (md != TTUF_MODE_8B) {
//JH     c = c & 0177;
//JH     if (md == TTUF_MODE_UC) {
//JH         if (islower (c)) c = toupper (c);
//JH         if ((mode & TTUF_KSR) && (c >= 0140))
//JH             return -1;
//JH         }
//JH     if (((md == TTUF_MODE_UC) || (md == TTUF_MODE_7P)) &&
//JH         ((c == 0177) ||
//JH          ((c < 040) && !((sim_tt_pchar >> c) & 1))))
//JH         return -1;
//JH     }
//JH else c = c & 0377;
//JH return c;
//JH }
//JH
//JH /* VMS routines, from Ben Thomas, with fixes from Robert Alan Byer */
//JH
//JH #if defined (VMS)
//JH
//JH #if defined(__VAX)
//JH #define sys$assign SYS$ASSIGN
//JH #define sys$qiow SYS$QIOW
//JH #endif
//JH
//JH #include <descrip.h>
//JH #include <ttdef.h>
//JH #include <tt2def.h>
//JH #include <iodef.h>
//JH #include <ssdef.h>
//JH #include <starlet.h>
//JH #include <unistd.h>
//JH
//JH #define EFN 0
//JH uint32 tty_chan = 0;
//JH
//JH typedef struct {
//JH     unsigned short sense_count;
//JH     unsigned char sense_first_char;
//JH     unsigned char sense_reserved;
//JH     unsigned int stat;
//JH     unsigned int stat2; } SENSE_BUF;
//JH
//JH typedef struct {
//JH     unsigned short status;
//JH     unsigned short count;
//JH     unsigned int dev_status; } IOSB;
//JH
//JH SENSE_BUF cmd_mode = { 0 };
//JH SENSE_BUF run_mode = { 0 };
//JH
//JH t_stat sim_ttinit (void)
//JH {
//JH unsigned int status;
//JH IOSB iosb;
//JH $DESCRIPTOR (terminal_device, "tt");
//JH
//JH status = sys$assign (&terminal_device, &tty_chan, 0, 0);
//JH if (status != SS$_NORMAL) return SCPE_TTIERR;
//JH status = sys$qiow (EFN, tty_chan, IO$_SENSEMODE, &iosb, 0, 0,
//JH     &cmd_mode, sizeof (cmd_mode), 0, 0, 0, 0);
//JH if ((status != SS$_NORMAL) || (iosb.status != SS$_NORMAL)) return SCPE_TTIERR;
//JH run_mode = cmd_mode;
//JH run_mode.stat = cmd_mode.stat | TT$M_NOECHO & ~(TT$M_HOSTSYNC | TT$M_TTSYNC);
//JH run_mode.stat2 = cmd_mode.stat2 | TT2$M_PASTHRU;
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttrun (void)
//JH {
//JH unsigned int status;
//JH IOSB iosb;
//JH
//JH status = sys$qiow (EFN, tty_chan, IO$_SETMODE, &iosb, 0, 0,
//JH     &run_mode, sizeof (run_mode), 0, 0, 0, 0);
//JH if ((status != SS$_NORMAL) || (iosb.status != SS$_NORMAL)) return SCPE_TTIERR;
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttcmd (void)
//JH {
//JH unsigned int status;
//JH IOSB iosb;
//JH
//JH status = sys$qiow (EFN, tty_chan, IO$_SETMODE, &iosb, 0, 0,
//JH     &cmd_mode, sizeof (cmd_mode), 0, 0, 0, 0);
//JH if ((status != SS$_NORMAL) || (iosb.status != SS$_NORMAL)) return SCPE_TTIERR;
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttclose (void)
//JH {
//JH return sim_ttcmd ();
//JH }
//JH
//JH t_stat sim_os_poll_kbd (void)
//JH {
//JH unsigned int status, term[2];
//JH unsigned char buf[4];
//JH IOSB iosb;
//JH SENSE_BUF sense;
//JH
//JH term[0] = 0; term[1] = 0;
//JH status = sys$qiow (EFN, tty_chan, IO$_SENSEMODE | IO$M_TYPEAHDCNT, &iosb,
//JH     0, 0, &sense, 8, 0, term, 0, 0);
//JH if ((status != SS$_NORMAL) || (iosb.status != SS$_NORMAL)) return SCPE_TTIERR;
//JH if (sense.sense_count == 0) return SCPE_OK;
//JH term[0] = 0; term[1] = 0;
//JH status = sys$qiow (EFN, tty_chan,
//JH     IO$_READLBLK | IO$M_NOECHO | IO$M_NOFILTR | IO$M_TIMED | IO$M_TRMNOECHO,
//JH     &iosb, 0, 0, buf, 1, 0, term, 0, 0);
//JH if ((status != SS$_NORMAL) || (iosb.status != SS$_NORMAL)) return SCPE_OK;
//JH if (buf[0] == sim_int_char) return SCPE_STOP;
//JH if (sim_brk_char && (buf[0] == sim_brk_char)) return SCPE_BREAK;
//JH return (buf[0] | SCPE_KFLAG);
//JH }
//JH
//JH t_stat sim_os_putchar (int32 out)
//JH {
//JH unsigned int status;
//JH char c;
//JH IOSB iosb;
//JH
//JH c = out;
//JH status = sys$qiow (EFN, tty_chan, IO$_WRITELBLK | IO$M_NOFORMAT,
//JH     &iosb, 0, 0, &c, 1, 0, 0, 0, 0);
//JH if ((status != SS$_NORMAL) || (iosb.status != SS$_NORMAL)) return SCPE_TTOERR;
//JH return SCPE_OK;
//JH }
//JH
//JH /* Win32 routines */
//JH
//JH #elif defined (_WIN32)
//JH
//JH #include <conio.h>
//JH #include <fcntl.h>
//JH #include <io.h>
//JH #include <windows.h>
//JH #define RAW_MODE 0
//JH static HANDLE std_input;
//JH static DWORD saved_mode;
//JH
//JH t_stat sim_ttinit (void)
//JH {
//JH std_input = GetStdHandle (STD_INPUT_HANDLE);
//JH if ((std_input == INVALID_HANDLE_VALUE) ||
//JH     !GetConsoleMode (std_input, &saved_mode)) return SCPE_TTYERR;
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttrun (void)
//JH {
//JH if (!GetConsoleMode(std_input, &saved_mode) ||
//JH     !SetConsoleMode(std_input, RAW_MODE)) return SCPE_TTYERR;
//JH if (sim_log) {
//JH     fflush (sim_log);
//JH     _setmode (_fileno (sim_log), _O_BINARY);
//JH     }
//JH SetThreadPriority (GetCurrentThread(), THREAD_PRIORITY_BELOW_NORMAL);
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttcmd (void)
//JH {
//JH if (sim_log) {
//JH     fflush (sim_log);
//JH     _setmode (_fileno (sim_log), _O_TEXT);
//JH     }
//JH SetThreadPriority (GetCurrentThread(), THREAD_PRIORITY_NORMAL);
//JH if (!SetConsoleMode(std_input, saved_mode)) return SCPE_TTYERR;
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttclose (void)
//JH {
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_os_poll_kbd (void)
//JH {
//JH int c;
//JH
//JH if (!_kbhit ()) return SCPE_OK;
//JH c = _getch ();
//JH if ((c & 0177) == sim_del_char) c = 0177;
//JH if ((c & 0177) == sim_int_char) return SCPE_STOP;
//JH if (sim_brk_char && ((c & 0177) == sim_brk_char)) return SCPE_BREAK;
//JH return c | SCPE_KFLAG;
//JH }
//JH
//JH t_stat sim_os_putchar (int32 c)
//JH {
//JH if (c != 0177) _putch (c);
//JH return SCPE_OK;
//JH }
//JH
//JH /* OS/2 routines, from Bruce Ray and Holger Veit */
//JH
//JH #elif defined (__OS2__)
//JH
//JH #include <conio.h>
//JH
//JH t_stat sim_ttinit (void)
//JH {
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttrun (void)
//JH {
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttcmd (void)
//JH {
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttclose (void)
//JH {
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_os_poll_kbd (void)
//JH {
//JH int c;
//JH
//JH #if defined (__EMX__)
//JH switch (c = _read_kbd(0,0,0)) {                         /* EMX has _read_kbd */
//JH
//JH     case -1:                                            /* no char*/
//JH         return SCPE_OK;
//JH
//JH     case 0:                                             /* char pending */
//JH         c = _read_kbd(0,1,0);
//JH         break;
//JH
//JH     default:                                            /* got char */
//JH         break;
//JH         }
//JH #else
//JH if (!kbhit ()) return SCPE_OK;
//JH c = getch();
//JH #endif
//JH if ((c & 0177) == sim_del_char) c = 0177;
//JH if ((c & 0177) == sim_int_char) return SCPE_STOP;
//JH if (sim_brk_char && ((c & 0177) == sim_brk_char)) return SCPE_BREAK;
//JH return c | SCPE_KFLAG;
//JH }
//JH
//JH t_stat sim_os_putchar (int32 c)
//JH {
//JH if (c != 0177) {
//JH #if defined (__EMX__)
//JH     putchar (c);
//JH #else
//JH     putch (c);
//JH #endif
//JH     fflush (stdout);
//JH     }
//JH return SCPE_OK;
//JH }
//JH
//JH /* Metrowerks CodeWarrior Macintosh routines, from Louis Chretien and
//JH    Peter Schorn */
//JH
//JH #elif defined (__MWERKS__) && defined (macintosh)
//JH
//JH #include <console.h>
//JH #include <Mactypes.h>
//JH #include <string.h>
//JH #include <sioux.h>
//JH #include <unistd.h>
//JH #include <siouxglobals.h>
//JH #include <Traps.h>
//JH #include <LowMem.h>
//JH
//JH /* function prototypes */
//JH
//JH Boolean SIOUXIsAppWindow(WindowPtr window);
//JH void SIOUXDoMenuChoice(long menuValue);
//JH void SIOUXUpdateMenuItems(void);
//JH void SIOUXUpdateScrollbar(void);
//JH int ps_kbhit(void);
//JH int ps_getch(void);
//JH
//JH extern char sim_name[];
//JH extern pSIOUXWin SIOUXTextWindow;
//JH static CursHandle iBeamCursorH = NULL;                  /* contains the iBeamCursor */
//JH
//JH static void updateCursor(void) {
//JH     WindowPtr window;
//JH     window = FrontWindow();
//JH     if (SIOUXIsAppWindow(window)) {
//JH         GrafPtr savePort;
//JH         Point localMouse;
//JH         GetPort(&savePort);
//JH         SetPort(window);
//JH #if TARGET_API_MAC_CARBON
//JH         GetGlobalMouse(&localMouse);
//JH #else
//JH         localMouse = LMGetMouseLocation();
//JH #endif
//JH         GlobalToLocal(&localMouse);
//JH         if (PtInRect(localMouse, &(*SIOUXTextWindow->edit)->viewRect) && iBeamCursorH) {
//JH             SetCursor(*iBeamCursorH);
//JH         }
//JH         else {
//JH             SetCursor(&qd.arrow);
//JH         }
//JH         TEIdle(SIOUXTextWindow->edit);
//JH         SetPort(savePort);
//JH     }
//JH     else {
//JH         SetCursor(&qd.arrow);
//JH         TEIdle(SIOUXTextWindow->edit);
//JH     }
//JH     return;
//JH }
//JH
//JH int ps_kbhit(void) {
//JH     EventRecord event;
//JH     int c;
//JH     updateCursor();
//JH     SIOUXUpdateScrollbar();
//JH     while (GetNextEvent(updateMask | osMask | mDownMask | mUpMask | activMask |
//JH              highLevelEventMask | diskEvt, &event)) {
//JH         SIOUXHandleOneEvent(&event);
//JH     }
//JH     if (SIOUXQuitting) {
//JH         exit(1);
//JH     }
//JH     if (EventAvail(keyDownMask,&event)) {
//JH         c = event.message&charCodeMask;
//JH         if ((event.modifiers & cmdKey) && (c > 0x20)) {
//JH             GetNextEvent(keyDownMask, &event);
//JH             SIOUXHandleOneEvent(&event);
//JH             if (SIOUXQuitting) {
//JH                 exit(1);
//JH             }
//JH             return false;
//JH         }
//JH         return true;
//JH     }
//JH     else {
//JH         return false;
//JH     }
//JH }
//JH
//JH int ps_getch(void) {
//JH     int c;
//JH     EventRecord event;
//JH     fflush(stdout);
//JH     updateCursor();
//JH     while(!GetNextEvent(keyDownMask,&event)) {
//JH         if (GetNextEvent(updateMask | osMask | mDownMask | mUpMask | activMask |
//JH              highLevelEventMask | diskEvt, &event)) {
//JH             SIOUXUpdateScrollbar();
//JH             SIOUXHandleOneEvent(&event);
//JH         }
//JH     }
//JH     if (SIOUXQuitting) {
//JH         exit(1);
//JH     }
//JH     c = event.message&charCodeMask;
//JH     if ((event.modifiers & cmdKey) && (c > 0x20)) {
//JH         SIOUXUpdateMenuItems();
//JH         SIOUXDoMenuChoice(MenuKey(c));
//JH     }
//JH     if (SIOUXQuitting) {
//JH         exit(1);
//JH     }
//JH    return c;
//JH }
//JH
//JH /* Note that this only works if the call to sim_ttinit comes before any output to the console */
//JH
//JH t_stat sim_ttinit (void) {
//JH     int i;
//JH     /* this blank will later be replaced by the number of characters */
//JH     char title[50] = " ";
//JH     unsigned char ptitle[50];
//JH     SIOUXSettings.autocloseonquit       = TRUE;
//JH     SIOUXSettings.asktosaveonclose = FALSE;
//JH     SIOUXSettings.showstatusline = FALSE;
//JH     SIOUXSettings.columns = 80;
//JH     SIOUXSettings.rows = 40;
//JH     SIOUXSettings.toppixel = 42;
//JH     SIOUXSettings.leftpixel     = 6;
//JH     iBeamCursorH = GetCursor(iBeamCursor);
//JH     strcat(title, sim_name);
//JH     strcat(title, " Simulator");
//JH     title[0] = strlen(title) - 1;                       /* Pascal string done */
//JH     for (i = 0; i <= title[0]; i++) {                   /* copy to unsigned char */
//JH         ptitle[i] = title[i];
//JH     }
//JH     SIOUXSetTitle(ptitle);
//JH     return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttrun (void)
//JH {
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttcmd (void)
//JH {
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttclose (void)
//JH {
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_os_poll_kbd (void)
//JH {
//JH int c;
//JH
//JH if (!ps_kbhit ()) return SCPE_OK;
//JH c = ps_getch();
//JH if ((c & 0177) == sim_del_char) c = 0177;
//JH if ((c & 0177) == sim_int_char) return SCPE_STOP;
//JH if (sim_brk_char && ((c & 0177) == sim_brk_char)) return SCPE_BREAK;
//JH return c | SCPE_KFLAG;
//JH }
//JH
//JH t_stat sim_os_putchar (int32 c)
//JH {
//JH if (c != 0177) {
//JH     putchar (c);
//JH     fflush (stdout);
//JH     }
//JH return SCPE_OK;
//JH }
//JH
//JH /* BSD UNIX routines */
//JH
//JH #elif defined (BSDTTY)
//JH
//JH #include <sgtty.h>
//JH #include <fcntl.h>
//JH #include <unistd.h>
//JH
//JH struct sgttyb cmdtty,runtty;                            /* V6/V7 stty data */
//JH struct tchars cmdtchars,runtchars;                      /* V7 editing */
//JH struct ltchars cmdltchars,runltchars;                   /* 4.2 BSD editing */
//JH int cmdfl,runfl;                                        /* TTY flags */
//JH
//JH t_stat sim_ttinit (void)
//JH {
//JH cmdfl = fcntl (0, F_GETFL, 0);                          /* get old flags  and status */
//JH runfl = cmdfl | FNDELAY;
//JH if (ioctl (0, TIOCGETP, &cmdtty) < 0) return SCPE_TTIERR;
//JH if (ioctl (0, TIOCGETC, &cmdtchars) < 0) return SCPE_TTIERR;
//JH if (ioctl (0, TIOCGLTC, &cmdltchars) < 0) return SCPE_TTIERR;
//JH runtty = cmdtty;                                        /* initial run state */
//JH runtty.sg_flags = cmdtty.sg_flags & ~(ECHO|CRMOD) | CBREAK;
//JH runtchars.t_intrc = sim_int_char;                       /* interrupt */
//JH runtchars.t_quitc = 0xFF;                               /* no quit */
//JH runtchars.t_startc = 0xFF;                              /* no host sync */
//JH runtchars.t_stopc = 0xFF;
//JH runtchars.t_eofc = 0xFF;
//JH runtchars.t_brkc = 0xFF;
//JH runltchars.t_suspc = 0xFF;                              /* no specials of any kind */
//JH runltchars.t_dsuspc = 0xFF;
//JH runltchars.t_rprntc = 0xFF;
//JH runltchars.t_flushc = 0xFF;
//JH runltchars.t_werasc = 0xFF;
//JH runltchars.t_lnextc = 0xFF;
//JH return SCPE_OK;                                         /* return success */
//JH }
//JH
//JH t_stat sim_ttrun (void)
//JH {
//JH runtchars.t_intrc = sim_int_char;                       /* in case changed */
//JH fcntl (0, F_SETFL, runfl);                              /* non-block mode */
//JH if (ioctl (0, TIOCSETP, &runtty) < 0) return SCPE_TTIERR;
//JH if (ioctl (0, TIOCSETC, &runtchars) < 0) return SCPE_TTIERR;
//JH if (ioctl (0, TIOCSLTC, &runltchars) < 0) return SCPE_TTIERR;
//JH nice (10);                                              /* lower priority */
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttcmd (void)
//JH {
//JH nice (-10);                                             /* restore priority */
//JH fcntl (0, F_SETFL, cmdfl);                              /* block mode */
//JH if (ioctl (0, TIOCSETP, &cmdtty) < 0) return SCPE_TTIERR;
//JH if (ioctl (0, TIOCSETC, &cmdtchars) < 0) return SCPE_TTIERR;
//JH if (ioctl (0, TIOCSLTC, &cmdltchars) < 0) return SCPE_TTIERR;
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttclose (void)
//JH {
//JH return sim_ttcmd ();
//JH }
//JH
//JH t_stat sim_os_poll_kbd (void)
//JH {
//JH int status;
//JH unsigned char buf[1];
//JH
//JH status = read (0, buf, 1);
//JH if (status != 1) return SCPE_OK;
//JH if (sim_brk_char && (buf[0] == sim_brk_char)) return SCPE_BREAK;
//JH else return (buf[0] | SCPE_KFLAG);
//JH }
//JH
//JH t_stat sim_os_putchar (int32 out)
//JH {
//JH char c;
//JH
//JH c = out;
//JH write (1, &c, 1);
//JH return SCPE_OK;
//JH }
//JH
//JH /* POSIX UNIX routines, from Leendert Van Doorn */
//JH
//JH #else
//JH
//JH #include <termios.h>
//JH #include <unistd.h>
//JH
//JH struct termios cmdtty, runtty;
//JH static int prior_norm = 1;
//JH
//JH t_stat sim_ttinit (void)
//JH {
//JH if (!isatty (fileno (stdin))) return SCPE_OK;           /* skip if !tty */
//JH if (tcgetattr (0, &cmdtty) < 0) return SCPE_TTIERR;     /* get old flags */
//JH runtty = cmdtty;
//JH runtty.c_lflag = runtty.c_lflag & ~(ECHO | ICANON);     /* no echo or edit */
//JH runtty.c_oflag = runtty.c_oflag & ~OPOST;               /* no output edit */
//JH runtty.c_iflag = runtty.c_iflag & ~ICRNL;               /* no cr conversion */
//JH runtty.c_cc[VINTR] = sim_int_char;                      /* interrupt */
//JH runtty.c_cc[VQUIT] = 0;                                 /* no quit */
//JH runtty.c_cc[VERASE] = 0;
//JH runtty.c_cc[VKILL] = 0;
//JH runtty.c_cc[VEOF] = 0;
//JH runtty.c_cc[VEOL] = 0;
//JH runtty.c_cc[VSTART] = 0;                                /* no host sync */
//JH runtty.c_cc[VSUSP] = 0;
//JH runtty.c_cc[VSTOP] = 0;
//JH #if defined (VREPRINT)
//JH runtty.c_cc[VREPRINT] = 0;                              /* no specials */
//JH #endif
//JH #if defined (VDISCARD)
//JH runtty.c_cc[VDISCARD] = 0;
//JH #endif
//JH #if defined (VWERASE)
//JH runtty.c_cc[VWERASE] = 0;
//JH #endif
//JH #if defined (VLNEXT)
//JH runtty.c_cc[VLNEXT] = 0;
//JH #endif
//JH runtty.c_cc[VMIN] = 0;                                  /* no waiting */
//JH runtty.c_cc[VTIME] = 0;
//JH #if defined (VDSUSP)
//JH runtty.c_cc[VDSUSP] = 0;
//JH #endif
//JH #if defined (VSTATUS)
//JH runtty.c_cc[VSTATUS] = 0;
//JH #endif
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttrun (void)
//JH {
//JH if (!isatty (fileno (stdin))) return SCPE_OK;           /* skip if !tty */
//JH runtty.c_cc[VINTR] = sim_int_char;                      /* in case changed */
//JH if (tcsetattr (0, TCSAFLUSH, &runtty) < 0) return SCPE_TTIERR;
//JH if (prior_norm) {                                       /* at normal pri? */
//JH     errno =     0;
//JH     nice (10);                                          /* try to lower pri */
//JH     prior_norm = errno;                                 /* if no error, done */
//JH     }
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttcmd (void)
//JH {
//JH if (!isatty (fileno (stdin))) return SCPE_OK;           /* skip if !tty */
//JH if (!prior_norm) {                                      /* priority down? */
//JH     errno =     0;
//JH     nice (-10);                                         /* try to raise pri */
//JH     prior_norm = (errno == 0);                          /* if no error, done */
//JH     }
//JH if (tcsetattr (0, TCSAFLUSH, &cmdtty) < 0) return SCPE_TTIERR;
//JH return SCPE_OK;
//JH }
//JH
//JH t_stat sim_ttclose (void)
//JH {
//JH return sim_ttcmd ();
//JH }
//JH
//JH t_stat sim_os_poll_kbd (void)
//JH {
//JH int status;
//JH unsigned char buf[1];
//JH
//JH status = read (0, buf, 1);
//JH if (status != 1) return SCPE_OK;
//JH if (sim_brk_char && (buf[0] == sim_brk_char)) return SCPE_BREAK;
//JH else return (buf[0] | SCPE_KFLAG);
//JH }
//JH
//JH t_stat sim_os_putchar (int32 out)
//JH {
//JH char c;
//JH
//JH c = out;
//JH write (1, &c, 1);
//JH return SCPE_OK;
//JH }
//JH
//JH #endif
